// SPDX-License-Identifier: GPL-2.0

#include <linux/ctype.h>
#include <linux/delay.h>
#include <linux/device.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/media-bus-format.h>
#include <linux/module.h>
#include <linux/of.h>
#include <linux/of_graph.h>
#include <linux/regulator/consumer.h>
#include <linux/types.h>
#include <linux/uaccess.h>
#include <linux/version.h>
#include <linux/v4l2-mediabus.h>
#include <media/v4l2-async.h>
#include <media/v4l2-ctrls.h>
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 13, 0)
#include <media/v4l2-fwnode.h>
#else
#include <media/v4l2-of.h>
#endif
#include <media/v4l2-subdev.h>
#include "basler-camera-driver.h"
#include "nvidia-support.h"
#include "device.h"
#include "register-io.h"
#include "memoryXX.h"

#ifdef NVIDIA_BUILD
#define DRIVER_VERSION	"1.8.1-nvidia"
#else
#define DRIVER_VERSION	"1.8.1"
#endif
#define SENSOR_NAME	"basler-camera"


static int basler_camera_s_ctrl(struct v4l2_ctrl *ctrl);
static int basler_camera_g_volatile_ctrl(struct v4l2_ctrl *ctrl);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(6, 1, 0)
static int basler_camera_validate(const struct v4l2_ctrl *ctrl,
				  union v4l2_ctrl_ptr ptr);
static bool basler_camera_equal(const struct v4l2_ctrl *ctrl,
				union v4l2_ctrl_ptr ptr1,
				union v4l2_ctrl_ptr ptr2);
#else
static int basler_camera_validate(const struct v4l2_ctrl *ctrl, u32 idx,
				  union v4l2_ctrl_ptr ptr);
static bool basler_camera_equal(const struct v4l2_ctrl *ctrl, u32 idx,
				union v4l2_ctrl_ptr ptr1,
				union v4l2_ctrl_ptr ptr2);
#endif
static void basler_camera_init(const struct v4l2_ctrl *ctrl, u32 idx,
			       union v4l2_ctrl_ptr ptr);
static int basler_camera_read_device_information_cache(
	struct basler_camera_dev *sensor,
	struct basler_device_information **pp_device_information);


#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 20, 0) // since 4.20.0-rc1
int basler_retrieve_csi_information(struct basler_camera_dev *sensor,
					   struct basler_csi_information *bci)
{
	struct device *dev = &sensor->i2c_client->dev;
	struct device_node *ep;
	struct v4l2_fwnode_endpoint bus_cfg = {
		.bus_type = V4L2_MBUS_CSI2_DPHY,
	};
	int ret = 0;
	unsigned short i;

	/* We need a function that searches for the device that holds
	 * the csi-2 bus information. For now we put the bus information
	 * also into the sensor endpoint itself.
	 */
	ep = of_graph_get_next_endpoint(dev->of_node, NULL);
	if (!ep) {
		dev_err(dev, "missing endpoint node\n");
		return -ENODEV;
	}

	ret = v4l2_fwnode_endpoint_alloc_parse(of_fwnode_handle(ep), &bus_cfg);
	of_node_put(ep);
	if (ret) {
		dev_err(dev, "failed to parse endpoint\n");
		goto done;
	}

	if (bus_cfg.bus_type != V4L2_MBUS_CSI2_DPHY ||
	    bus_cfg.bus.mipi_csi2.num_data_lanes == 0 ||
	    bus_cfg.nr_of_link_frequencies == 0) {
		dev_err(dev, "missing CSI-2 properties in endpoint\n");
		ret = -ENODATA;
		goto done;
	}

	bci->max_lane_frequency = bus_cfg.link_frequencies[0];
	bci->lane_count = bus_cfg.bus.mipi_csi2.num_data_lanes;
	for (i = 0; i < bus_cfg.bus.mipi_csi2.num_data_lanes; ++i)
		bci->lane_assignment[i] = bus_cfg.bus.mipi_csi2.data_lanes[i];

done:
	v4l2_fwnode_endpoint_free(&bus_cfg);
	return ret;
}
#else
int basler_retrieve_csi_information(struct basler_camera_dev *sensor,
					   struct basler_csi_information *bci)
{
	struct device *dev = &sensor->i2c_client->dev;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 13, 0)
	struct v4l2_fwnode_endpoint *endpoint;
#else
	struct v4l2_of_endpoint *endpoint;
#endif
	struct device_node *ep;
	int ret = 0;
	unsigned short i;

	/* We need a function that searches for the device that holds
	 * the csi-2 bus information. For now we put the bus information
	 * also into the sensor endpoint itself.
	 */
	ep = of_graph_get_next_endpoint(dev->of_node, NULL);
	if (!ep) {
		dev_err(dev, "missing endpoint node\n");
		return -ENODEV;
	}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 13, 0)
	endpoint = v4l2_fwnode_endpoint_alloc_parse(of_fwnode_handle(ep));
	of_node_put(ep);
#else
	endpoint = v4l2_of_alloc_parse_endpoint(ep);
#endif
	if (IS_ERR(endpoint)) {
		dev_err(dev, "failed to parse endpoint\n");
		return PTR_ERR(endpoint);
	}

	if (endpoint->bus_type != V4L2_MBUS_CSI2 ||
	    endpoint->bus.mipi_csi2.num_data_lanes == 0 ||
	    endpoint->nr_of_link_frequencies == 0) {
		dev_err(dev, "missing CSI-2 properties in endpoint\n");
		ret = -ENODATA;
		goto done;
	}

	bci->max_lane_frequency = endpoint->link_frequencies[0];
	bci->lane_count = endpoint->bus.mipi_csi2.num_data_lanes;
	for (i = 0; i < endpoint->bus.mipi_csi2.num_data_lanes; ++i)
		bci->lane_assignment[i] = endpoint->bus.mipi_csi2.data_lanes[i];

done:
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 13, 0)
	v4l2_fwnode_endpoint_free(endpoint);
#else
	v4l2_of_free_endpoint(endpoint);
#endif
	return ret;
}
#endif

static inline struct basler_camera_dev *
to_basler_camera_dev(struct v4l2_subdev *subdev)
{
#ifdef NVIDIA_BUILD
	return to_camera_common_data(subdev->dev)->priv;
#else
	return container_of(subdev, struct basler_camera_dev, sd_);
#endif
}

static inline struct v4l2_subdev *ctrl_to_sd(struct v4l2_ctrl *ctrl)
{
	return container_of(ctrl->handler, struct basler_camera_dev,
			     ctrl_handler)
			->sd;
}

/**
 * basler_camera_set_fmt - set format of the camera
 *
 * Note: We just store the format so we can answer the `g_fmt` op.
 *       Proper configuration of the camera will be done from user space
 *
 * Returns zero for pad 0 and -EINVAL for all other pads
 */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 14, 0)
static int basler_camera_set_fmt(struct v4l2_subdev *sd,
				 struct v4l2_subdev_state *state,
				 struct v4l2_subdev_format *format)
#else
static int basler_camera_set_fmt(struct v4l2_subdev *sd,
				 struct v4l2_subdev_pad_config *cfg,
				 struct v4l2_subdev_format *format)
#endif
{
	struct basler_camera_dev *sensor = to_basler_camera_dev(sd);

	if (format->pad)
		return -EINVAL;

	mutex_lock(&sensor->lock);

	sensor->format = format->format;

	mutex_unlock(&sensor->lock);

	return 0;
}

/**
 * basler_camera_get_fmt - Get current format
 *
 * Note: We just return the format from the last `s_fmt` invocation.
 *       Proper configuration of the camera will be done from user space
 *
 * Returns zero for pad 0 and -EINVAL for all other pads
 */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 14, 0)
static int basler_camera_get_fmt(struct v4l2_subdev *sd,
				 struct v4l2_subdev_state *state,
				 struct v4l2_subdev_format *format)
#else
static int basler_camera_get_fmt(struct v4l2_subdev *sd,
				 struct v4l2_subdev_pad_config *cfg,
				 struct v4l2_subdev_format *format)
#endif
{
	struct basler_camera_dev *sensor = to_basler_camera_dev(sd);

	if (format->pad)
		return -EINVAL;

	mutex_lock(&sensor->lock);

	format->format = sensor->format;

	mutex_unlock(&sensor->lock);

	return 0;
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 14, 0)
static int basler_camera_enum_mbus_code(struct v4l2_subdev *sd,
					struct v4l2_subdev_state *state,
					struct v4l2_subdev_mbus_code_enum *code)
#else
static int basler_camera_enum_mbus_code(struct v4l2_subdev *sd,
					struct v4l2_subdev_pad_config *cfg,
					struct v4l2_subdev_mbus_code_enum *code)
#endif
{
	static uint16_t supported_datatypes[] = {
		MEDIA_BUS_FMT_RGB888_1X24, // RGB888
		MEDIA_BUS_FMT_UYVY8_1X16, // YUV422 8-bit
		MEDIA_BUS_FMT_YUYV8_2X8, // YUV422 8-bit
		MEDIA_BUS_FMT_SRGGB8_1X8, // RAW8
		MEDIA_BUS_FMT_SRGGB10_1X10, // RAW10
		MEDIA_BUS_FMT_SRGGB12_1X12 // RAW12
	};

	if (code->index >= ARRAY_SIZE(supported_datatypes))
		return -EINVAL;

	code->code = supported_datatypes[code->index];

	pr_debug("%s: %s() index=%u -> code=%04X\n", sd->name, __func__,
		 code->index, code->code);

	return 0;
}

/*
 * basler_camera_enum_framesizes() - returns the supported frame sizes by index.
 * @sd  the v4l2 subdevice
 * @cfg the pad configuration
 * @fse the frame size returned.
 *
 * Return 0 on success and an error otherwise.
 */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 14, 0)
static int
basler_camera_enum_framesizes(struct v4l2_subdev *sd,
			      struct v4l2_subdev_state *state,
			      struct v4l2_subdev_frame_size_enum *fse)
#else
static int
basler_camera_enum_framesizes(struct v4l2_subdev *sd,
			      struct v4l2_subdev_pad_config *cfg,
			      struct v4l2_subdev_frame_size_enum *fse)
#endif
{
	pr_debug("%s: %s(...)\n", sd->name, __func__);

#if 0
	if (fse->code != V4L2_PIX_FMT_UYVY &&
		fse->code != V4L2_PIX_FMT_SRGGB8 &&
		fse->code != V4L2_PIX_FMT_ABGR32 &&
		fse->code != V4L2_PIX_FMT_GREY)
		return -EINVAL;
#endif
	if (fse->index >= 1)
		return -EINVAL;

	fse->min_width = 4;
	fse->min_height = 4;
	fse->max_width = 5000;
	fse->max_height = 4000;

	pr_debug("%s: %s(...) => %d/%d .. %d/%d\n", sd->name, __func__,
		 fse->min_width, fse->min_height, fse->max_width,
		 fse->max_height);

	return 0;
}

/**
 * basler_camera_s_stream - start camera streaming
 *
 * Note: Will be done in user space
 *
 * Returns always zero
 */
static int basler_camera_s_stream(struct v4l2_subdev *sd, int enable)
{
	return 0;
}

static int basler_camera_s_power(struct v4l2_subdev *sd, int on)
{
	struct basler_camera_dev *sensor = to_basler_camera_dev(sd);

	mutex_lock(&sensor->lock);

	/* Update the power count. */
	sensor->power_count += on ? 1 : -1;
	WARN_ON(sensor->power_count < 0);

	mutex_unlock(&sensor->lock);

	return 0;
}

static long basler_camera_priv_ioctl(struct v4l2_subdev *sd, unsigned int cmd, void *arg)
{
	struct basler_camera_dev *sensor = to_basler_camera_dev(sd);
	int ret;

	switch (cmd) {
	case BASLER_IOC_G_INTERFACE_VERSION:
	{
		__u32 __user *uptr = (__u32 __user *)arg;

		__u32 version = (((__u32)BASLER_INTERFACE_VERSION_MAJOR) << 16) | BASLER_INTERFACE_VERSION_MINOR;
		ret = put_user(version, uptr);
		if (ret)
			return -EFAULT;

		return 0;
	}

	case BASLER_IOC_READ_REGISTER: {
		struct register_access __user *p_ra_user =
			(struct register_access __user *)arg;
		struct register_access ra_tmp;

		ret = basler_ra_copy_from_userspace(&ra_tmp, p_ra_user);
		if (ret < 0)
			return ret;

		ret = basler_rw_memory_16bit(sensor, &ra_tmp);
		if (ret < 0)
			return ret;

		ret = copy_to_user(p_ra_user->data, ra_tmp.data,
				   ra_tmp.data_size);
		if (ret)
			return -EFAULT;

		return 0;
	}

	case BASLER_IOC_READ_REGISTER_32: {
		struct register_access_32 __user *p_ra_user =
			(struct register_access_32 __user *)arg;
		struct register_access_32 ra_tmp;

		ret = basler_ra_32_copy_from_userspace(&ra_tmp, p_ra_user);
		if (ret < 0)
			return ret;

		ret = basler_rw_memory_32bit(sensor, &ra_tmp);
		if (ret < 0)
			return ret;

		ret = copy_to_user(p_ra_user->data, ra_tmp.data,
				   ra_tmp.data_size);
		if (ret)
			return -EFAULT;

		return 0;
	}

	case BASLER_IOC_WRITE_REGISTER: {
		struct register_access __user *p_ra_user =
			(struct register_access __user *)arg;
		struct register_access ra_tmp;

		ret = basler_ra_copy_from_userspace(&ra_tmp, p_ra_user);
		if (ret < 0)
			return ret;

		ret = basler_rw_memory_16bit(sensor, &ra_tmp);
		if (ret < 0)
			return ret;

		return 0;
	}

	case BASLER_IOC_WRITE_REGISTER_32: {
		struct register_access_32 __user *p_ra_user =
			(struct register_access_32 __user *)arg;
		struct register_access_32 ra_tmp;

		ret = basler_ra_32_copy_from_userspace(&ra_tmp, p_ra_user);
		if (ret < 0)
			return ret;

		ret = basler_rw_memory_32bit(sensor, &ra_tmp);
		if (ret < 0)
			return ret;

		return 0;
	}

	case BASLER_IOC_G_DEVICE_INFORMATION:
	{
		struct basler_device_information __user *uptr = (struct basler_device_information __user *)arg;
		struct basler_device_information *p_bdi = NULL;

		ret = basler_camera_read_device_information_cache(sensor, &p_bdi);
		if (ret < 0) {
			return ret;
		}

		ret = copy_to_user(uptr, p_bdi, sizeof(struct basler_device_information));
		if (ret)
			return -EFAULT;

		return 0;
	}

	case BASLER_IOC_G_CSI_INFORMATION:
	{
		struct basler_csi_information __user *uptr = (struct basler_csi_information __user *)arg;
		struct basler_csi_information csi_info;

		ret = basler_retrieve_csi_information(sensor, &csi_info);
		if (ret < 0)
			return ret;

		ret = copy_to_user(uptr, &csi_info, sizeof(struct basler_csi_information));
		if (ret)
			return -EFAULT;

		return 0;
	}

	case BASLER_IOC_G_I2C_INFORMATION:
	{
		struct basler_i2c_information __user *uptr = (struct basler_i2c_information __user *)arg;
		struct basler_i2c_information i2c_info;

		i2c_info.adapter_id = i2c_adapter_id(sensor->i2c_client->adapter);
		i2c_info.device_address = sensor->i2c_client->addr;

		ret = copy_to_user(uptr, &i2c_info, sizeof(struct basler_i2c_information));
		if (ret)
			return -EFAULT;

		return 0;
	}

	default:
		break;
	}

	return -EINVAL;
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 10, 0)
static int basler_camera_g_mbus_config(struct v4l2_subdev *sd,
				       unsigned int pad,
				       struct v4l2_mbus_config *cfg)
#else
static int basler_camera_g_mbus_config(struct v4l2_subdev *sd,
				       struct v4l2_mbus_config *cfg)
#endif
{

#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 20, 0)
	cfg->type = V4L2_MBUS_CSI2_DPHY;
#else
	cfg->type = V4L2_MBUS_CSI2;
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 18, 0)
	cfg->bus.mipi_csi2.num_data_lanes = 2;
	cfg->bus.mipi_csi2.flags = 0;
#else
	cfg->flags = V4L2_MBUS_CSI2_CONTINUOUS_CLOCK | V4L2_MBUS_CSI2_CHANNEL_0;
	// cfg->flags |= V4L2_MBUS_CSI2_1_LANE;
	cfg->flags |= V4L2_MBUS_CSI2_2_LANE;
	//cfg->flags |= V4L2_MBUS_CSI2_3_LANE;
	//cfg->flags |= V4L2_MBUS_CSI2_4_LANE;
#endif

	return 0;
}

static const struct v4l2_subdev_core_ops basler_camera_core_ops = {
	.s_power = basler_camera_s_power,
	.ioctl = basler_camera_priv_ioctl,
};

static const struct v4l2_subdev_video_ops basler_camera_video_ops = {
	.s_stream = basler_camera_s_stream,
#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 10, 0)
	.g_mbus_config = basler_camera_g_mbus_config,
#endif
};

static const struct v4l2_subdev_pad_ops basler_camera_pad_ops = {
	.set_fmt = basler_camera_set_fmt,
	.get_fmt = basler_camera_get_fmt,
	.enum_mbus_code = basler_camera_enum_mbus_code,
	.enum_frame_size = basler_camera_enum_framesizes,
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 10, 0)
	.get_mbus_config = basler_camera_g_mbus_config,
#endif
};

static const struct v4l2_subdev_ops basler_camera_subdev_ops = {
	.core = &basler_camera_core_ops,
	.video = &basler_camera_video_ops,
	.pad = &basler_camera_pad_ops,
};

static const struct v4l2_ctrl_ops basler_camera_ctrl_ops = {
	.g_volatile_ctrl = basler_camera_g_volatile_ctrl,
	.s_ctrl = basler_camera_s_ctrl,
};

static const struct v4l2_ctrl_type_ops basler_camera_ctrl_type_ops = {
	.validate = basler_camera_validate,
	.init = basler_camera_init,
	.equal = basler_camera_equal,
};

/**
 * basler_camera_validate
 *
 * Note: Not needed by access-register control
 *
 * Returns always zero
 */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(6, 1, 0)
static int basler_camera_validate(const struct v4l2_ctrl *ctrl,
				  union v4l2_ctrl_ptr ptr)
#else
static int basler_camera_validate(const struct v4l2_ctrl *ctrl, u32 idx,
				  union v4l2_ctrl_ptr ptr)
#endif
{
	return 0;
}

/**
 * basler_camera_init
 *
 * Note: Not needed by access-register control
 *
 */
static void basler_camera_init(const struct v4l2_ctrl *ctrl, u32 idx,
			       union v4l2_ctrl_ptr ptr)
{
}

/**
 * basler_camera_equal
 *
 * Note: Not needed by access-register control
 *
 * Returns always zero
 */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(6, 1, 0)
static bool basler_camera_equal(const struct v4l2_ctrl *ctrl,
				union v4l2_ctrl_ptr ptr1,
				union v4l2_ctrl_ptr ptr2)
#else
static bool basler_camera_equal(const struct v4l2_ctrl *ctrl, u32 idx,
				union v4l2_ctrl_ptr ptr1,
				union v4l2_ctrl_ptr ptr2)
#endif
{
	return 0;
}

/* We need to define a new V4L2_CTRL_TYPE value that does not collide
 * with an existing one.
 *
 * The basic V4L2_CTRL_TYPE_XXX values are defined in linux/videodev2.h
 * as enum v4l2_ctrl_type, but several other headers extend these
 * definitions by adding more values through #defines. There seems to be
 * no central coordination facility to avoid collisions when doing so.
 * Therefore, the best we can do is selecting some obscure random value,
 * cross our fingers and hope for the best.
 */
#define V4L2_CTRL_TYPE_UNIQUE	(0x370de8ca)

static const struct v4l2_ctrl_config ctrl_access_register = {
	.ops = &basler_camera_ctrl_ops,
	.type_ops = &basler_camera_ctrl_type_ops,
	.id = V4L2_CID_BASLER_ACCESS_REGISTER,
	.name = "basler-access-register",
	.type = V4L2_CTRL_TYPE_UNIQUE,
	.flags = V4L2_CTRL_FLAG_EXECUTE_ON_WRITE | V4L2_CTRL_FLAG_VOLATILE,
	.step = 1,
	.dims = { 1 },
	.elem_size = sizeof(struct register_access),
};

static const struct v4l2_ctrl_config ctrl_access_register_32 = {
	.ops = &basler_camera_ctrl_ops,
	.type_ops = &basler_camera_ctrl_type_ops,
	.id = V4L2_CID_BASLER_ACCESS_REGISTER_32,
	.name = "basler-access-register-32",
	.type = V4L2_CTRL_TYPE_UNIQUE,
	.flags = V4L2_CTRL_FLAG_EXECUTE_ON_WRITE | V4L2_CTRL_FLAG_VOLATILE,
	.step = 1,
	.dims = { 1 },
	.elem_size = sizeof(struct register_access_32),
};

static const struct v4l2_ctrl_config ctrl_basler_device_information = {
	.ops = &basler_camera_ctrl_ops,
	.type_ops = &basler_camera_ctrl_type_ops,
	.id = V4L2_CID_BASLER_DEVICE_INFORMATION,
	.name = "basler-device-information",
	.type = V4L2_CTRL_TYPE_UNIQUE,
	.flags = V4L2_CTRL_FLAG_READ_ONLY | V4L2_CTRL_FLAG_VOLATILE,
	.step = 1,
	.dims = { 1 },
	.elem_size = sizeof(struct basler_device_information),
};

static const struct v4l2_ctrl_config ctrl_basler_interface_version = {
	.ops = &basler_camera_ctrl_ops,
	.type_ops = &basler_camera_ctrl_type_ops,
	.id = V4L2_CID_BASLER_INTERFACE_VERSION,
	.name = "basler-interface-version",
	.type = V4L2_CTRL_TYPE_INTEGER,
	.flags = V4L2_CTRL_FLAG_READ_ONLY,
	.min = 0x0,
	.max = 0xffffffff,
	.def = (BASLER_INTERFACE_VERSION_MAJOR << 16) |
	       BASLER_INTERFACE_VERSION_MINOR,
	.step = 1,
};

static const struct v4l2_ctrl_config ctrl_basler_csi_information = {
	.ops = &basler_camera_ctrl_ops,
	.type_ops = &basler_camera_ctrl_type_ops,
	.id = V4L2_CID_BASLER_CSI_INFORMATION,
	.name = "basler-csi-information",
	.type = V4L2_CTRL_TYPE_UNIQUE,
	.flags = V4L2_CTRL_FLAG_READ_ONLY | V4L2_CTRL_FLAG_VOLATILE,
	.step = 1,
	.dims = { 1 },
	.elem_size = sizeof(struct basler_csi_information),
};

static const struct v4l2_ctrl_config ctrl_basler_i2c_information = {
	.ops = &basler_camera_ctrl_ops,
	.type_ops = &basler_camera_ctrl_type_ops,
	.id = V4L2_CID_BASLER_I2C_INFORMATION,
	.name = "basler-i2c-information",
	.type = V4L2_CTRL_TYPE_UNIQUE,
	.flags = V4L2_CTRL_FLAG_READ_ONLY | V4L2_CTRL_FLAG_VOLATILE,
	.step = 1,
	.dims = { 1 },
	.elem_size = sizeof(struct basler_i2c_information),
};

static int basler_camera_s_ctrl(struct v4l2_ctrl *ctrl)
{
	struct v4l2_subdev *sd = ctrl_to_sd(ctrl);
	struct basler_camera_dev *sensor = to_basler_camera_dev(sd);
	int ret;
	struct register_access *fp_ra_16_new;
	struct register_access_32 *fp_ra_32_new;

	switch (ctrl->id) {
	case V4L2_CID_BASLER_ACCESS_REGISTER:
		if (ctrl->elem_size != sizeof(struct register_access)) {
			dev_err(sd->dev,
				"%s: size mismatch V4L2_CID_BASLER_ACCESS_REGISTER: elem = %u, expected = %zu",
				__func__, ctrl->elem_size,
				sizeof(struct register_access));
			return -ENOMEM;
		}
		fp_ra_16_new = (struct register_access *)ctrl->p_new.p;

		if (I2CREAD == (fp_ra_16_new->command | I2CREAD)) {
			//save the memory address and size for later read
			// - we have to save the pair (address + size) for each process,
			// reason; the accesses (_store() + _load() ) are not serialized.
			// - in user mode, serialization is performed only through the threads.
			// - we use a simple implementation without dynamic memory allocations.
			ret = process_local_ra_storage_store(
				&sensor->ra_read_container,
				fp_ra_16_new->address, fp_ra_16_new->data_size);
			return ret;
		} else if (I2CWRITE == (fp_ra_16_new->command | I2CWRITE)) {
			ret = basler_rw_memory_16bit(sensor, fp_ra_16_new);
			if (ret < 0)
				return ret;
		} else
			return -EPERM;

		break;

	case V4L2_CID_BASLER_ACCESS_REGISTER_32:
		if (ctrl->elem_size != sizeof(struct register_access_32)) {
			dev_err(sd->dev,
				"%s: size mismatch V4L2_CID_BASLER_ACCESS_REGISTER: elem = %u, expected = %zu",
				__func__, ctrl->elem_size,
				sizeof(struct register_access_32));
			return -ENOMEM;
		}
		fp_ra_32_new = (struct register_access_32 *)ctrl->p_new.p;

		if (I2CREAD == (fp_ra_32_new->command | I2CREAD)) {
			ret = process_local_ra_storage_store(
				&sensor->ra_read_container,
				fp_ra_32_new->address, fp_ra_32_new->data_size);
			return ret;
		} else if (I2CWRITE == (fp_ra_32_new->command | I2CWRITE)) {
			ret = basler_rw_memory_32bit(sensor, fp_ra_32_new);
			if (ret < 0)
				return ret;
		} else
			return -EPERM;

		break;

	default:
		return -EINVAL;
	}

	return 0;
}

static int
basler_camera_g_ctrl_access_register(struct device *dev,
				     struct basler_camera_dev *sensor,
				     const struct v4l2_ctrl *ctrl)
{
	u32 tmp_address, tmp_data_size;
	struct register_access *fp_ra_16_new = NULL;
	struct register_access_32 *fp_ra_32_new = NULL;
	int ret = 0;

	switch (ctrl->id) {
	case V4L2_CID_BASLER_ACCESS_REGISTER:
		if (ctrl->elem_size != sizeof(struct register_access)) {
			dev_err(dev,
				"%s: size mismatch V4L2_CID_BASLER_ACCESS_REGISTER: elem = %u, expected = %zu",
				__func__, ctrl->elem_size,
				sizeof(struct register_access));
			return -ENOMEM;
		}

		fp_ra_16_new = (struct register_access *)ctrl->p_new.p;
		ret = process_local_ra_storage_load_and_free(
			&sensor->ra_read_container, &tmp_address,
			&tmp_data_size);
		if (ret < 0)
			return ret;
		fp_ra_16_new->address = tmp_address;
		fp_ra_16_new->data_size = tmp_data_size;
		fp_ra_16_new->command = I2CREAD;
		ret = basler_rw_memory_16bit(sensor, fp_ra_16_new);
		break;

	case V4L2_CID_BASLER_ACCESS_REGISTER_32:
		if (ctrl->elem_size != sizeof(struct register_access_32)) {
			dev_err(dev,
				"%s: size mismatch V4L2_CID_BASLER_ACCESS_REGISTER: elem = %u, expected = %zu",
				__func__, ctrl->elem_size,
				sizeof(struct register_access_32));
			return -ENOMEM;
		}

		fp_ra_32_new = (struct register_access_32 *)ctrl->p_new.p;
		ret = process_local_ra_storage_load_and_free(
			&sensor->ra_read_container, &tmp_address,
			&tmp_data_size);
		if (ret < 0)
			return ret;
		fp_ra_32_new->address = tmp_address;
		fp_ra_32_new->data_size = tmp_data_size;
		fp_ra_32_new->command = I2CREAD;
		ret = basler_rw_memory_32bit(sensor, fp_ra_32_new);
		break;

	default:
		return -EINVAL;
	}

	if (ret < 0)
		return ret;

	if (ret == 0)
		return -EIO;

	return 0;
}

static int
basler_camera_g_ctrl_device_information(struct device *dev,
					struct basler_camera_dev *sensor,
					const struct v4l2_ctrl *ctrl)
{
	struct basler_device_information *l_bdi = NULL;
	struct basler_device_information *p_bdi = NULL;
	int ret = 0;

	if (ctrl->elem_size != sizeof(struct basler_device_information)) {
		dev_err(dev,
			"%s: size mismatch V4L2_CID_BASLER_DEVICE_INFORMATION: elem = %u, expected = %zu",
			__func__, ctrl->elem_size,
			sizeof(struct basler_device_information));
		return -ENOMEM;
	}

	l_bdi = (struct basler_device_information *)ctrl->p_new.p;

	ret = basler_camera_read_device_information_cache(sensor, &p_bdi);
	if (ret < 0) {
		return ret;
	}
	memcpy(l_bdi, p_bdi, sizeof(struct basler_device_information));

	return 0;
}

static int
basler_camera_g_ctrl_csi_information(struct device *dev,
				     struct basler_camera_dev *sensor,
				     const struct v4l2_ctrl *ctrl)
{
	struct basler_csi_information *l_bci = NULL;
	int ret;

	if (ctrl->elem_size != sizeof(struct basler_csi_information)) {
		dev_err(dev,
			"%s: size mismatch V4L2_CID_BASLER_CSI_INFORMATION: elem = %u, expected = %zu",
			__func__, ctrl->elem_size,
			sizeof(struct basler_csi_information));
		return -ENOMEM;
	}

	l_bci = (struct basler_csi_information *)ctrl->p_new.p;
	ret = basler_retrieve_csi_information(sensor, l_bci);
	if (ret < 0)
		return ret;

	return 0;
}

static int
basler_camera_g_ctrl_i2c_information(struct device *dev,
				     struct basler_camera_dev *sensor,
				     const struct v4l2_ctrl *ctrl)
{
	struct basler_i2c_information *l_bi2c = NULL;

	if (ctrl->elem_size != sizeof(struct basler_i2c_information)) {
		dev_err(dev,
			"%s: size mismatch V4L2_CID_BASLER_I2C_INFORMATION: elem = %u, expected = %zu",
			__func__, ctrl->elem_size,
			sizeof(struct basler_i2c_information));
		return -ENOMEM;
	}

	l_bi2c = (struct basler_i2c_information *)ctrl->p_new.p;
	l_bi2c->adapter_id = i2c_adapter_id(sensor->i2c_client->adapter);
	l_bi2c->device_address = sensor->i2c_client->addr;

	return 0;
}

static int basler_camera_g_volatile_ctrl(struct v4l2_ctrl *ctrl)
{
	struct v4l2_subdev *sd = ctrl_to_sd(ctrl);
	struct device *dev = sd->dev;
	struct basler_camera_dev *sensor = to_basler_camera_dev(sd);

	switch (ctrl->id) {
	case V4L2_CID_BASLER_ACCESS_REGISTER:
	case V4L2_CID_BASLER_ACCESS_REGISTER_32:
		return basler_camera_g_ctrl_access_register(dev, sensor, ctrl);
	case V4L2_CID_BASLER_DEVICE_INFORMATION:
		return basler_camera_g_ctrl_device_information(dev, sensor, ctrl);
	case V4L2_CID_BASLER_CSI_INFORMATION:
		return basler_camera_g_ctrl_csi_information(dev, sensor, ctrl);
	case V4L2_CID_BASLER_I2C_INFORMATION:
		return basler_camera_g_ctrl_i2c_information(dev, sensor, ctrl);
	default:
		break;
	}

	return -EINVAL;
}

/**
 * basler_camera_link_setup
 *
 * Note: Function is needed by imx8qm
 *
 * Returns always zero
 */
static int basler_camera_link_setup(struct media_entity *entity,
				    const struct media_pad *local,
				    const struct media_pad *remote, u32 flags)
{
	return 0;
}

static const struct media_entity_operations basler_camera_sd_media_ops = {
	.link_setup = basler_camera_link_setup,
};

static int basler_camera_alloc_control(struct v4l2_ctrl_handler *hdl,
				       const struct v4l2_ctrl_config *cfg,
				       struct device *dev)
{
	struct v4l2_ctrl *ctrl;
	int ret = 0;

	ctrl = v4l2_ctrl_new_custom(hdl, cfg, NULL);
	if (ctrl == NULL) {
		dev_err(dev,
			"Registering ctrl %s failed: %d\n",
			cfg->name,
			hdl->error);
		ret = hdl->error;
	}

	return ret;
}

static int basler_camera_init_controls(struct basler_camera_dev *sensor)
{
	struct v4l2_ctrl_handler *hdl = &sensor->ctrl_handler;

	v4l2_ctrl_handler_init(hdl, 32);

	/* we can use our own mutex for the ctrl lock */
	hdl->lock = &sensor->lock;

	if (basler_camera_alloc_control(hdl, &ctrl_access_register, &sensor->i2c_client->dev) ||
	    basler_camera_alloc_control(hdl, &ctrl_access_register_32, &sensor->i2c_client->dev) ||
	    basler_camera_alloc_control(hdl, &ctrl_basler_device_information, &sensor->i2c_client->dev) ||
	    basler_camera_alloc_control(hdl, &ctrl_basler_interface_version, &sensor->i2c_client->dev) ||
	    basler_camera_alloc_control(hdl, &ctrl_basler_csi_information, &sensor->i2c_client->dev) ||
	    basler_camera_alloc_control(hdl, &ctrl_basler_i2c_information, &sensor->i2c_client->dev)) {
		/* free up all controls allocated so far */
		v4l2_ctrl_handler_free(hdl);
		dev_dbg(sensor->sd->v4l2_dev->dev, "%s: ctrl setup error.\n",
			__func__);
		return hdl->error;
	}

	sensor->sd->ctrl_handler = hdl;
	return 0;
}

static int basler_camera_read_device_information_cache(
	struct basler_camera_dev *sensor,
	struct basler_device_information **pp_device_information)
{
	int ret = 0;

	*pp_device_information = &sensor->device_information;

	if (sensor->device_information_cached) {
		return 0;
	}

	// during this 'enumeration' call we update and cache the capabilities
	if ( basler_probe_32bit_capability(sensor->i2c_client) >= 0)
		atomic_set(&sensor->memory_address_size, BASLER_MEMROY_ADDRESS_SIZE_32BIT);
	else
		atomic_set(&sensor->memory_address_size, BASLER_MEMROY_ADDRESS_SIZE_16BIT);
	smp_wmb();
	pr_debug("used device address width %d[bytes]\n",
		 atomic_read(&sensor->memory_address_size));

	// but we do not cache the device information,
	// because we need it only during this 'enumeration' call
	ret = basler_retrieve_device_information(sensor,
						 &sensor->device_information);
	if (ret) {
		pr_debug("basler_retrieve_device_information() failed: %d\n",
			 ret);
		return ret;
	}

	return 0;
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(6, 3, 0)
static int basler_camera_probe(struct i2c_client *client)
#else
static int basler_camera_probe(struct i2c_client *client,
			       const struct i2c_device_id *id)
#endif
{
	struct device *dev = &client->dev;
	struct basler_camera_dev *sensor;
	int ret;

	dev_info(dev, "Basler Camera Driver (v" DRIVER_VERSION ") start probing\n");

	if (!i2c_check_functionality(client->adapter, I2C_FUNC_I2C)) {
		dev_err(dev, "I2C_FUNC_I2C not supported\n");
		return -ENODEV;
	}

	sensor = devm_kzalloc(dev, sizeof(*sensor), GFP_KERNEL);
	if (!sensor)
		return -ENOMEM;

	/* VDD: 5V camera supply */
	sensor->vdd = devm_regulator_get_optional(dev, "vdd");
	if (IS_ERR(sensor->vdd)) {
		if (PTR_ERR(sensor->vdd) != -ENODEV) {
			dev_err(dev, "failed to get VDD regulator\n");
			return PTR_ERR(sensor->vdd);
		}
		dev_dbg(dev, "no VDD regulator found\n");
		sensor->vdd = NULL;
	}

	if (sensor->vdd != NULL) {
		/* Turn on camera supply and leave it on */
		ret = regulator_enable(sensor->vdd);
		if (ret) {
			dev_err(dev, "failed to enable VDD regulator\n");
			return ret;
		}
	}

	/*
	 * Give the camera time to power up and initialize its I2C interface.
	 * The duration is deliberately chosen much longer than the actual start-up
	 * time of the currently supported cameras to account for future models and
	 * to error on the safe side since it's not critical
	 */
	msleep(1000);

	sensor->i2c_client = client;

#ifdef NVIDIA_BUILD
	basler_camera_set_default_format_for_nvidia(sensor);

	ret = basler_set_common_data_for_nvidia(dev, sensor, sensor->s_data);
	if (ret) {
		dev_err(dev, "basler_set_common_data_for_nvidia() failed: %d\n", ret);
		goto reg_off;
	}
#else
	sensor->sd = &sensor->sd_;
#endif

	v4l2_i2c_subdev_init(sensor->sd, client, &basler_camera_subdev_ops);

	atomic_set(&sensor->memory_address_size,
		   BASLER_MEMROY_ADDRESS_SIZE_UNKNOWN);
	sensor->device_information_cached = false;
	if (of_property_read_bool(dev->of_node, "skip_probing")) {
		dev_dbg(dev, "property: 'skip_probing' found\n");
	} else {
		if (basler_probe_32bit_capability(client) >= 0)
			atomic_set(&sensor->memory_address_size,
				   BASLER_MEMROY_ADDRESS_SIZE_32BIT);
		else
			atomic_set(&sensor->memory_address_size,
				   BASLER_MEMROY_ADDRESS_SIZE_16BIT);
		dev_dbg(dev, "used device address width %d[bytes]\n",
			atomic_read(&sensor->memory_address_size));

		ret = basler_retrieve_device_information(
			sensor, &sensor->device_information);
		if (ret) {
			dev_dbg(dev,
				"basler_retrieve_device_information() failed: %d\n",
				ret);
			goto reg_off;
		}
		sensor->device_information_cached = true;
	}

	sensor->sd->flags |= V4L2_SUBDEV_FL_HAS_DEVNODE;
	sensor->pad.flags = MEDIA_PAD_FL_SOURCE;
	sensor->sd->entity.function = MEDIA_ENT_F_CAM_SENSOR;
	sensor->sd->entity.ops = &basler_camera_sd_media_ops;
	ret = media_entity_pads_init(&sensor->sd->entity, 1, &sensor->pad);
	if (ret) {
		dev_dbg(dev, "media_entity_pads_init() failed: %d\n", ret);
		goto reg_off;
	}

	process_local_ra_storage_init(&sensor->ra_read_container);
	mutex_init(&sensor->lock);

	ret = basler_camera_init_controls(sensor);
	if (ret) {
		dev_dbg(dev,
			"basler_camera_init_controls() failed: %d\n",
			ret);
		goto entity_cleanup;
	}

	ret = v4l2_async_register_subdev(sensor->sd);
	if (ret) {
		dev_dbg(dev,
			"v4l2_async_register_subdev() failed: %d\n",
			ret);
		goto handler_cleanup;
	}

	dev_info(dev, "camera probed successfully\n");
	return 0;

handler_cleanup:
	sensor->sd->ctrl_handler = NULL;
	v4l2_ctrl_handler_free(&sensor->ctrl_handler);
entity_cleanup:
	mutex_destroy(&sensor->lock);
	process_local_ra_storage_release(&sensor->ra_read_container);
	media_entity_cleanup(&sensor->sd->entity);
reg_off:
	if (sensor->vdd != NULL) {
		regulator_disable(sensor->vdd);
	}

	return ret;
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(6, 1, 0)
static void basler_camera_remove(struct i2c_client *client)
#else
static int basler_camera_remove(struct i2c_client *client)
#endif
{
	struct v4l2_subdev *sd = i2c_get_clientdata(client);
	struct basler_camera_dev *sensor = to_basler_camera_dev(sd);

	v4l2_async_unregister_subdev(sensor->sd);
	sensor->sd->ctrl_handler = NULL;
	v4l2_ctrl_handler_free(&sensor->ctrl_handler);
	mutex_destroy(&sensor->lock);
	process_local_ra_storage_release(&sensor->ra_read_container);
	media_entity_cleanup(&sensor->sd->entity);

	if (sensor->vdd != NULL) {
		regulator_disable(sensor->vdd);
	}
#if LINUX_VERSION_CODE < KERNEL_VERSION(6, 1, 0)
	return 0;
#endif
}

static const struct i2c_device_id basler_camera_id[] = {
	{ "basler-camera", 0 },
	{},
};
MODULE_DEVICE_TABLE(i2c, basler_camera_id);

static const struct of_device_id basler_camera_dt_ids[] = {
	{ .compatible = "basler,basler-camera" },
	{ /* sentinel */ }
};
MODULE_DEVICE_TABLE(of, basler_camera_dt_ids);

static struct i2c_driver basler_camera_i2c_driver = {
	.driver = {
			.name = "basler-camera",
			.of_match_table = basler_camera_dt_ids,
		},
	.id_table = basler_camera_id,
	.probe = basler_camera_probe,
	.remove = basler_camera_remove,
};

module_i2c_driver(basler_camera_i2c_driver);

MODULE_DESCRIPTION("Basler camera subdev driver");
MODULE_AUTHOR("Basler AG <support.europe@baslerweb.com>");
MODULE_LICENSE("GPL");
MODULE_VERSION(DRIVER_VERSION);
