// SPDX-License-Identifier: GPL-2.0

#include <linux/kernel.h>
#include <linux/uaccess.h>
#include <linux/version.h>

#include "memoryXX.h"
#include "basler-camera-driver.h"
#include "device.h"

static int basler_read_memory_16bit(struct i2c_client *client,
				    struct register_access *ra_p);

static int basler_write_memory_16bit(struct i2c_client *client,
				     struct register_access *ra_p);

/**
 * copy the user register_access struct
 * Only the elements that are needed for the desired command(read/write) are transferred.
 * In addition, only the used elements in the payload memory are copied.
 * @return < 0 for any failure
 */
int basler_ra_copy_from_userspace(struct register_access *ra_dst,
				  struct register_access const __user *ra_src)
{
#if (LINUX_VERSION_CODE < KERNEL_VERSION(5, 0, 0))
	if (!access_ok(VERIFY_READ, ra_src, sizeof(struct register_access)))
		return -EFAULT;
#else
	if (!access_ok(ra_src, sizeof(struct register_access)))
		return -EFAULT;
#endif

	if (__get_user(ra_dst->address, &ra_src->address))
		return -EFAULT;
	if (__get_user(ra_dst->data_size, &ra_src->data_size))
		return -EFAULT;
	if (__get_user(ra_dst->command, &ra_src->command))
		return -EFAULT;

	if (ra_dst->command == I2CWRITE) {
		if (ra_dst->data_size > sizeof(ra_dst->data))
			return -EINVAL;

		if (__copy_from_user(ra_dst->data, ra_src->data,
				     ra_dst->data_size))
			return -EFAULT;
	}

	return 0;
}

int basler_rw_memory_16bit(struct basler_camera_dev *sensor,
			   struct register_access *ra_p)
{
	int ret;
	if (ra_p->data_size > sizeof(ra_p->data)) {
		dev_err(&sensor->i2c_client->dev,
			"i2c r/w array too big, max allowed %zu, got %d\n",
			sizeof(ra_p->data), ra_p->data_size);
		return -EINVAL;
	}
	if (atomic_read(&sensor->memory_address_size) !=
	    BASLER_MEMROY_ADDRESS_SIZE_16BIT)
		return -ENODEV;

	//pr_debug("basler_rw_memory_16bit(CMD: 0x%X, size: %d, 0x%04x [address])\n",
	//	 ra_p->command, ra_p->data_size, ra_p->address);

	switch (ra_p->command) {
	case I2CREAD:
		ret = basler_read_memory_16bit(sensor->i2c_client, ra_p);
		break;

	case I2CWRITE:
		ret = basler_write_memory_16bit(sensor->i2c_client, ra_p);
		break;

	default:
		dev_err(&sensor->i2c_client->dev,
			"basler_rw_memory_16bit() unsupported memory access: 0x%X\n",
			ra_p->command);
		ret = -EINVAL;
	}

	if (ret < 0)
		ra_p->data_size = 0;
	else
		ra_p->data_size = ret;
	return ret;
}

static int basler_write_memory_16bit(struct i2c_client *client,
				     struct register_access *ra_p)
{
	int ret;
	u16 old_address = ra_p->address;
	const int bytes_to_send = ra_p->data_size + sizeof(ra_p->address);

	ra_p->address = cpu_to_be16(ra_p->address);
	ret = i2c_master_send(client, (u8 *)ra_p, bytes_to_send);

	ra_p->address = old_address;
	if (ret > 0 && ret != bytes_to_send)
		ret = -EIO;
	return ret;
}

static int basler_read_memory_16bit(struct i2c_client *client,
				    struct register_access *ra_p)
{
	__be16 register_address = cpu_to_be16(ra_p->address);
	return basler_read_memory_raw_16bit(client, ra_p->data, ra_p->data_size,
					    register_address);
}

int basler_read_memory_raw_16bit(struct i2c_client *client, u8 *buffer,
				u16 bytes_to_read, __be16 register_address)
{
	struct i2c_msg msgs[2] = {};
	int ret;
	msgs[0].addr = client->addr;
	msgs[0].flags = 0;
	msgs[0].buf = (u8 *)&register_address;
	msgs[0].len = sizeof(register_address);

	msgs[1].addr = client->addr;
	msgs[1].flags = I2C_M_RD;
	msgs[1].buf = buffer;
	msgs[1].len = bytes_to_read;

	ret = i2c_transfer(client->adapter, msgs, 2);
	if (ret < 0) {
		pr_err("i2c_transfer() failed: %d\n", ret);
		return ret;
	}

	if (ret != 2) {
		pr_err("i2c_transfer() incomplete");
		return -EIO;
	}

	return msgs[1].len;
}
