// SPDX-License-Identifier: GPL-2.0

#include <linux/kernel.h>
#include <linux/uaccess.h>
#include <linux/version.h>

#include "memoryXX.h"
#include "basler-camera-driver.h"
#include "device.h"

#define I2C_PROTOCOL_CRC_SIZE 1
#define I2C_PROTOCOL_STATE_SIZE 1
#define GENCP_STATE_SUCCESS 0x00
#define GENCP_BASLER_I2C_INVALID_HEADER 0xA1

struct __attribute__((packed)) i2c_protocol_header {
	__be32 register_address;
	__be16 payload_size;
};

struct __attribute__((packed)) i2c_protocol_write_buffer {
	struct i2c_protocol_header header;
	u8 payload_crc8[BASLER_REGISTER_ACCESS_BUF_SIZE + I2C_PROTOCOL_CRC_SIZE];
};

struct __attribute__((packed)) i2c_protocol_read_buffer {
	struct i2c_protocol_header header;
	u8 payload_status_crc8[BASLER_REGISTER_ACCESS_BUF_SIZE +
			       I2C_PROTOCOL_STATE_SIZE + I2C_PROTOCOL_CRC_SIZE];
};

struct __attribute__((packed)) i2c_protocol_result {
	u8 gencp_status_code;
	u8 crc8;
};

static int basler_write_memory_raw_32bit(struct i2c_client *client, u8 *buffer,
					 u16 bytes_to_write,
					 __be32 register_address);

static u8 basler_calculate_crc8(const u8 *buffer, u16 bytes_to_calculate);
static int basler_verify_response(const struct i2c_protocol_result *p_result,
				  u8 desired_crc8);
static int basler_transfer_i2c_messages(struct i2c_client *client,
					struct i2c_msg *p_msg, u8 num_messages);

/**
 * copy the user register_access_32 struct
 * Only the elements that are needed for the desired command(read/write) are transferred.
 * In addition, only the used elements in the payload memory are copied.
 * @return < 0 for any failure
 */
int basler_ra_32_copy_from_userspace(
	struct register_access_32 *ra_dst,
	struct register_access_32 const __user *ra_src)
{
#if (LINUX_VERSION_CODE < KERNEL_VERSION(5, 0, 0))
	if (!access_ok(VERIFY_READ, ra_src, sizeof(struct register_access)))
		return -EFAULT;
#else
	if (!access_ok(ra_src, sizeof(struct register_access)))
		return -EFAULT;
#endif

	if (__get_user(ra_dst->address, &ra_src->address))
		return -EFAULT;
	if (__get_user(ra_dst->data_size, &ra_src->data_size))
		return -EFAULT;
	if (__get_user(ra_dst->command, &ra_src->command))
		return -EFAULT;

	if (ra_dst->command == I2CWRITE) {
		if (ra_dst->data_size > sizeof(ra_dst->data))
			return -EINVAL;

		if (__copy_from_user(ra_dst->data, ra_src->data,
				     ra_dst->data_size))
			return -EFAULT;
	}

	return 0;
}

/**
 * try the 32bit communication with the given device
 * @return < 0 for any failure, during the probing
 */
int basler_probe_32bit_capability(struct i2c_client *client)
{
	const u8 write_buffer[2] = { 0, 0 };
	struct i2c_protocol_result result_buffer;
	struct i2c_msg msgs[2] = {};
	int ret;
	u8 desired_crc8;

	//Note
	// The following I2C access is interpreted differently depending on the target type.
	// - for 16bit communication;
	//  we address 0x0000 and read two byte from this
	// (ABRM GenCP version)
	// - for 32bit;
	//   the write, is too short/invalid for any 32bit transaction type,
	//   the read, is a valid status read transaction,
	//
	// If the GenCP version ever takes the value 0xA16E, the identification would fail.
	//
	msgs[0].addr = client->addr;
	msgs[0].flags = 0;
	msgs[0].buf = (u8 *)&write_buffer;
	msgs[0].len = sizeof(write_buffer);

	msgs[1].addr = client->addr;
	msgs[1].flags = I2C_M_RD;
	msgs[1].buf = (u8 *)&result_buffer;
	msgs[1].len = sizeof(result_buffer);

	ret = basler_transfer_i2c_messages(client, msgs, 2);
	if (ret < 0)
		return ret;
	desired_crc8 = basler_calculate_crc8((u8 *)&result_buffer,
					     sizeof(result_buffer) - 1);
	pr_debug(
		"basler_probe_32bit_capability(), state: 0x%02X, crc: 0x%02X, 0x%02X [is/should]\n",
		result_buffer.gencp_status_code, result_buffer.crc8,
		desired_crc8);
	if (desired_crc8 != result_buffer.crc8)
		return -ENODEV;
	if (result_buffer.gencp_status_code != GENCP_BASLER_I2C_INVALID_HEADER)
		return -ENODEV;

	return 0;
}

int basler_rw_memory_32bit(struct basler_camera_dev *sensor,
			   struct register_access_32 *ra_p)
{
	//Note:
	// a zero length write are used by the mcmif to query the (device) support for dword addressing

	int ret;
	__be32 register_address = cpu_to_be32(ra_p->address);
	if (ra_p->data_size > sizeof(ra_p->data)) {
		dev_err(&sensor->i2c_client->dev,
			"i2c r/w array too big, max allowed %zu, got %d\n",
			sizeof(ra_p->data), ra_p->data_size);
		return -EINVAL;
	}
	if (atomic_read(&sensor->memory_address_size) !=
	    BASLER_MEMROY_ADDRESS_SIZE_32BIT)
		return -ENODEV;
	if (ra_p->data_size == 0)
		return 0;
	if (ra_p->data_size > BASLER_REGISTER_ACCESS_BUF_SIZE)
		return -ENOMEM;

	//pr_debug("basler_rw_memory_32bit(CMD: 0x%X, size: %d, 0x%04x [address])\n",
	//	 ra_p->command, ra_p->data_size, ra_p->address);

	switch (ra_p->command) {
	case I2CREAD:
		ret = basler_read_memory_raw_32bit(sensor->i2c_client,
						   ra_p->data, ra_p->data_size,
						   register_address);
		break;

	case I2CWRITE:
		ret = basler_write_memory_raw_32bit(sensor->i2c_client,
						    ra_p->data, ra_p->data_size,
						    register_address);
		break;

	default:
		dev_err(&sensor->i2c_client->dev,
			"basler_rw_memory_32bit() unsupported memory access: 0x%X\n",
			ra_p->command);
		ret = -EINVAL;
	}

	if (ret < 0)
		ra_p->data_size = 0;
	else
		ra_p->data_size = ret;
	return ret;
}

int basler_read_memory_raw_32bit(struct i2c_client *client, u8 *buffer,
				 u16 bytes_to_read, __be32 register_address)
{
	struct i2c_protocol_read_buffer read_buffer;
	struct i2c_msg msgs[2] = {};
	int ret;
	u8 current_crc8;
	struct i2c_protocol_result *p_result;
	u16 used_buffer_size;

	// first we send/inform the slave about the memory address and size
	read_buffer.header.register_address = register_address;
	read_buffer.header.payload_size = cpu_to_be16(bytes_to_read);

	// combine both i2c transfers (header and payload+state+crc) via repeated start
	msgs[0].addr = client->addr;
	msgs[0].flags = 0;
	msgs[0].buf = (u8 *)&read_buffer;
	msgs[0].len = sizeof(read_buffer.header);

	msgs[1].addr = client->addr;
	msgs[1].flags = I2C_M_RD;
	msgs[1].buf = read_buffer.payload_status_crc8;
	msgs[1].len = bytes_to_read + sizeof(struct i2c_protocol_result);

	// send the paket and wait for response
	ret = basler_transfer_i2c_messages(client, msgs, 2);
	if (ret < 0)
		return ret;

	// verify the response
	used_buffer_size = sizeof(read_buffer.header) + bytes_to_read +
			   sizeof(struct i2c_protocol_result);
	current_crc8 =
		basler_calculate_crc8((u8 *)&read_buffer, used_buffer_size - 1);
	p_result =
		(struct i2c_protocol_result *)(read_buffer.payload_status_crc8 +
					       bytes_to_read);
	ret = basler_verify_response(p_result, current_crc8);
	if (ret < 0)
		return ret;

	memcpy(buffer, read_buffer.payload_status_crc8, bytes_to_read);
	return bytes_to_read;
}

static int basler_write_memory_raw_32bit(struct i2c_client *client, u8 *buffer,
					 u16 bytes_to_write,
					 __be32 register_address)
{
	struct i2c_protocol_write_buffer write_buffer;
	struct i2c_protocol_result result_buffer;
	struct i2c_msg msgs[2] = {};
	u16 used_write_buffer_size;
	int ret;
	u8 current_crc8;

	// fill the packet, the used length depends on the payload size
	used_write_buffer_size = sizeof(struct i2c_protocol_header) +
				 bytes_to_write + I2C_PROTOCOL_CRC_SIZE;
	write_buffer.header.register_address = register_address;
	write_buffer.header.payload_size = cpu_to_be16(bytes_to_write);
	memcpy(write_buffer.payload_crc8, buffer, bytes_to_write);
	write_buffer.payload_crc8[bytes_to_write] = basler_calculate_crc8(
		(u8 *)&write_buffer, used_write_buffer_size - 1);

	// combine both i2c transfers (header+payload and result) via repeated start
	msgs[0].addr = client->addr;
	msgs[0].flags = 0;
	msgs[0].buf = (u8 *)&write_buffer;
	msgs[0].len = used_write_buffer_size;

	msgs[1].addr = client->addr;
	msgs[1].flags = I2C_M_RD;
	msgs[1].buf = (u8 *)&result_buffer;
	msgs[1].len = sizeof(result_buffer);

	// send the paket and wait for response
	ret = basler_transfer_i2c_messages(client, msgs, 2);
	if (ret < 0)
		return ret;

	// verify the response
	current_crc8 = basler_calculate_crc8((u8 *)&result_buffer,
					     sizeof(result_buffer) - 1);
	return basler_verify_response(&result_buffer, current_crc8);
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 14, 0)
#define POLY (0x1070U << 3)
static u8 crc8(u16 data)
{
	int i;
	for (i = 0; i < 8; i++) {
		if (data & 0x8000)
			data = data ^ POLY;
		data = data << 1;
	}
	return (u8)(data >> 8);
}

static u8 i2c_smbus_pec(u8 crc, u8 *p, size_t count)
{
	size_t i;
	for (i = 0; i < count; i++)
		crc = crc8((crc ^ p[i]) << 8);
	return crc;
}
#endif

static u8 basler_calculate_crc8(const u8 *buffer, u16 bytes_to_calculate)
{
	//TODO for optimisation, we could use crc8_populate_msb() / crc8() which use a LUT
	return i2c_smbus_pec(0, (u8 *)buffer, bytes_to_calculate);
}

static int basler_verify_response(const struct i2c_protocol_result *p_result,
				  u8 desired_crc8)
{
	if (desired_crc8 != p_result->crc8) {
		pr_warn("CRC missmatch: is 0x%02X, should 0x%02X\n",
			p_result->crc8, desired_crc8);
		return -EIO;
	}
	if (p_result->gencp_status_code != GENCP_STATE_SUCCESS) {
		pr_warn("GenCP error: 0x%02X\n", p_result->gencp_status_code);
		return -EIO;
	}

	return 0;
}

static int basler_transfer_i2c_messages(struct i2c_client *client,
					struct i2c_msg *p_msg, u8 num_messages)
{
	int ret;
	ret = i2c_transfer(client->adapter, p_msg, num_messages);
	if (ret < 0) {
		pr_err("i2c_transfer() failed: %d\n", ret);
		return ret;
	}

	if (ret != num_messages) {
		pr_err("i2c_transfer() incomplete");
		return -EIO;
	}

	return 0;
}