// SPDX-License-Identifier: GPL-2.0


#include <linux/device.h>
#include <linux/of.h>
#include <linux/slab.h>
#include <linux/version.h>

#include "basler-camera-driver.h"
#include "nvidia-support.h"


#ifdef NVIDIA_BUILD
static const int basler_30fps[] = {
	30,
};

/*
 * WARNING: frmfmt ordering need to match mode definition in
 * device tree!
 */
static const struct camera_common_frmfmt basler_frmfmt[] = {
	/* size, framerates, num framerates, hdr, mode */
	{{1920, 1080}, basler_30fps, 1, 0, 0}
	/* Add modes with no device tree support after below */
};
#endif


#ifdef CONFIG_OF_DYNAMIC
static int get_num_of_sensor_modes_in_dt(struct device *dev)
{
	char temp_str[256];
	struct device_node *node = NULL;
	int num_modes = 0;
	int err = 0;

	while (true) {
		err = snprintf(temp_str, sizeof(temp_str), "mode%d", num_modes);
		if (err < 0)
			break;
		node = of_get_child_by_name(dev->of_node, temp_str);
		of_node_put(node);
		if (node == NULL)
			break;
		num_modes++;
	}
	dev_dbg(dev, "get_num_of_sensor_modes: %i\n", num_modes);
	return num_modes;
}

static bool node_has_property(struct device_node *node,
			      const char *property_name)
{
	const char *unused;
	if (of_property_read_string(node, property_name, &unused)) {
		return false;
	}
	return true;
}

static struct property *create_property(const char *name, const char *value)
{
	struct property *prop;
	prop = kzalloc(sizeof(*prop), GFP_KERNEL);
	if (!prop)
		return NULL;

	prop->name = kstrdup(name, GFP_KERNEL);
	prop->value = kstrdup(value, GFP_KERNEL);
	prop->length = strlen(value) + 1;
	if (!prop->name || !prop->value)
		goto err_free;

	/* mark the property as dynamic */
	of_property_set_flag(prop, OF_DYNAMIC);
	return prop;
err_free:
	kfree(prop->name);
	kfree(prop->value);
	kfree(prop);
	return NULL;
}

static int add_node_property_to_changeset(struct of_changeset *ocs,
					  struct device_node *node,
					  const char *prop_name,
					  const char *prop_value)
{
	struct property *prop;
	if (node_has_property(node, prop_name)) {
		return -EEXIST;
	}
	prop = create_property(prop_name, prop_value);
	if (!prop) {
		return -EINVAL;
	}
	return of_changeset_add_property(ocs, node, prop);
}

static int basler_set_devicetree_nodes_for_nvidia_modes(
	struct device *dev, struct basler_csi_information *csi_info)
{
	int ret = 0;
	struct device_node *node = NULL;
	struct of_changeset ocs;
	const int bits_per_pixel = 8;
	const int pixels_per_mipi_clock_tick = 2;
	// The pixel clock is needed to intialize the VI graph within the nvidia kernel.
	// So we can not set this value during runtime (for example set_fmt).
	// nvidia searches through all modes (we only have one) and takes the highest value.
	// we take an upper guess: Mipi lane clock * 2 * number of lanes / bits per pixel
	// for "bits per pixel" we take the smallest possible value which is 8.
	u64 pixel_clock = csi_info->max_lane_frequency *
			  pixels_per_mipi_clock_tick * csi_info->lane_count /
			  bits_per_pixel;
	char char_buffer[20];

	node = of_get_child_by_name(dev->of_node, "mode0");
	of_node_put(node);
	if (node == NULL) {
		dev_err(dev,
			"basler_set_devicetree_nodes_for_nvidia_modes() mode0 was not found: %d\n",
			ret);
		return -EINVAL;
	}

	of_changeset_init(&ocs);

	// here we write the properties that are mandatory for nvidia, but these are mostly values
	// that are used to calculate the MIPI clock rate and the pixel format, which we override.
	// So, we write them hard via code into device tree, to get rid of noise within our device
	// tree files
	// we don't override properties, so they still can be set in the device tree
	snprintf(char_buffer, sizeof(char_buffer), "%llu", pixel_clock);
	add_node_property_to_changeset(&ocs, node, "pix_clk_hz", char_buffer);
	snprintf(char_buffer, sizeof(char_buffer), "%u", csi_info->lane_count);
	add_node_property_to_changeset(&ocs, node, "num_lanes", char_buffer);
	// the rest are dummy values
	add_node_property_to_changeset(&ocs, node, "csi_pixel_bit_depth", "10");
	add_node_property_to_changeset(&ocs, node, "active_w", "100");
	add_node_property_to_changeset(&ocs, node, "active_h", "100");
	add_node_property_to_changeset(&ocs, node, "line_length", "100");
	add_node_property_to_changeset(&ocs, node, "mode_type", "bayer");
	add_node_property_to_changeset(&ocs, node, "pixel_phase", "bggr");
	add_node_property_to_changeset(&ocs, node, "phy_mode", "DPHY");

	ret = of_changeset_apply(&ocs);
	if (ret) {
		dev_err(dev, "of_changeset_apply() failed: %d\n", ret);
		goto changeset_cleanup;
	}

changeset_cleanup:
	of_changeset_destroy(&ocs);
	return ret;
}

#endif //CONFIG_OF_DYNAMIC

#ifdef NVIDIA_BUILD
int
basler_set_common_data_for_nvidia(struct device *dev,
				  struct basler_camera_dev *sensor,
				  struct camera_common_data *common_data)
{
	int ret = 0;
	u64 mipi_clock;
	int num_of_sensor_modes;
	struct basler_csi_information csi_info;
	ret = basler_retrieve_csi_information(sensor, &csi_info);
	if (ret) {
		dev_err(dev, "basler_retrieve_csi_information() failed: %d\n",
			ret);
		return ret;
	}
	num_of_sensor_modes = get_num_of_sensor_modes_in_dt(dev);
	if (num_of_sensor_modes != 1) {
		dev_err(dev, "We need exactly one sensor mode but found %d\n",
			num_of_sensor_modes);
		return -EINVAL;
	}
	common_data = devm_kzalloc(dev, sizeof(*common_data), GFP_KERNEL);

	common_data->priv = sensor;
	common_data->dev = dev;
	sensor->s_data = common_data;
	sensor->sd = &common_data->subdev;
	sensor->sd->dev = dev;

	common_data->def_clk_freq = csi_info.max_lane_frequency;
	// the rest are dummy values
	common_data->frmfmt = &basler_frmfmt[0];
	common_data->numfmts = ARRAY_SIZE(basler_frmfmt);
	common_data->def_mode = 0;
	common_data->def_width = 1920;
	common_data->def_height = 1080;
	common_data->fmt_width = common_data->def_width;
	common_data->fmt_height = common_data->def_height;

	ret = basler_set_devicetree_nodes_for_nvidia_modes(dev, &csi_info);
	if (ret) {
		dev_err(dev,
			"basler_set_devicetree_nodes_for_nvidia_modes() failed: %d\n",
			ret);
		return ret;
	}

	common_data->colorfmt =
		camera_common_find_datafmt(MEDIA_BUS_FMT_UYVY8_1X16);

	// <!> Attention <!>
	// inside the funtion
	//   camera_common_initialize(struct camera_common_data *s_data,
	//                                   const char *dev_name) {
	//     char debugfs_name[xx];
	//     ...
	//     sprintf(debugfs_name, "%s_%c", dev_name, s_data->csi_port + 'a');
	//     ...}
	//kernel_src/kernel/nvidia/drivers/media/platform/tegra/camera/camera_common.c
	// 35.1 <> debugfs_name[35]
	// 34.1 <> debugfs_name[10]
	// 32.7.1 <> debugfs_name[35]
	// 32.6.1 <> debugfs_name[10]
	// if the name is too long then it will couse an oops ...
	// "NX Boot failed" https://forums.developer.nvidia.com/t/nx-boot-failed/185041/8
	ret = camera_common_initialize(common_data, "basler");
	if (ret) {
		dev_err(dev, "camera_common_initialize() failed: %d\n", ret);
		return ret;
	}
	mipi_clock = csi_info.max_lane_frequency;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 10, 0)
	// the mipi clock rate is calculated within the camera_common_initialize function.
	// but we want to use the value specified in the device tree endpoint.
	common_data->sensor_props.sensor_modes[0]
		.signal_properties.mipi_clock.val = mipi_clock;
#else
	// mipi clock is based on the pixel clock.
	// nothing to do here
#endif
	if (csi_info.lane_count != common_data->sensor_props.sensor_modes[0]
					   .signal_properties.num_lanes) {
		dev_err(dev,
			"consistency check for lane count failed: in the device tree of this sensor are two different lane counts. Please check your device tree.\n");
		return -EINVAL;
	}
	return ret;
}

void
basler_camera_set_default_format_for_nvidia(struct basler_camera_dev *sensor)
{
	struct v4l2_mbus_framefmt *fmt;

	fmt = &sensor->format;
	fmt->code = MEDIA_BUS_FMT_YUYV8_1X16;
	fmt->colorspace = V4L2_COLORSPACE_SRGB;
	fmt->ycbcr_enc = V4L2_MAP_YCBCR_ENC_DEFAULT(fmt->colorspace);
	fmt->quantization = V4L2_MAP_QUANTIZATION_DEFAULT(true,
							  fmt->colorspace,
							  fmt->ycbcr_enc);
	fmt->xfer_func = V4L2_MAP_XFER_FUNC_DEFAULT(fmt->colorspace);
	fmt->width = 1920;
	fmt->height = 1080;
	fmt->field = V4L2_FIELD_NONE;
}

#endif //NVIDIA_BUILD

