// SPDX-License-Identifier: GPL-2.0

#include "process-local-ra-storage.h"

#include <linux/sched.h>

static void
remove_item_locked(struct process_local_ra_storage_container *p_storage,
		   unsigned int item_index_to_remove);
static void remove_items_for_pid_locked(
	struct process_local_ra_storage_container *p_storage, pid_t pid);
static int
find_free_index_locked(struct process_local_ra_storage_container *p_storage);
static int
find_oldest_item_locked(struct process_local_ra_storage_container *p_storage);
static int
find_item_for_pid(struct process_local_ra_storage_container *p_storage,
		  pid_t pid);
static s32 find_largest_order_number_locked(
	struct process_local_ra_storage_container *p_storage);
__attribute__((unused)) static void
dump_storage_locked(struct process_local_ra_storage_container *p_storage);

void process_local_ra_storage_init(
	struct process_local_ra_storage_container *p_storage)
{
	unsigned int i;
	for (i = 0; i < ARRAY_SIZE(p_storage->items); i++)
		p_storage->items[i].is_used = 0;
	mutex_init(&p_storage->lock);
}

void process_local_ra_storage_release(
	struct process_local_ra_storage_container *p_storage)
{
	mutex_destroy(&p_storage->lock);
}

int process_local_ra_storage_store(
	struct process_local_ra_storage_container *p_storage, u32 ra_address,
	u32 ra_data_size)
{
	int index;
	s32 max_order_number;
	const pid_t caller_pid = current->pid;

	if (p_storage == NULL)
		return -EINVAL;

	//pr_debug(
	//	"process_local_ra_storage_store(PID: %d, size: %d, address: 0x%04x)\n",
	//	caller_pid, ra_data_size, ra_address);

	mutex_lock(&p_storage->lock);
	//------------------------------------------------>

	// there should never be an entry for the PID, just to be on the safe side
	remove_items_for_pid_locked(p_storage, caller_pid);

	index = find_free_index_locked(p_storage);
	if (index < 0) {
		index = find_oldest_item_locked(p_storage);
		remove_item_locked(p_storage, index);
	}

	max_order_number = find_largest_order_number_locked(p_storage);
	p_storage->items[index].is_used = true;
	p_storage->items[index].order_number = max_order_number + 1;
	p_storage->items[index].item_owner = caller_pid;
	p_storage->items[index].address = ra_address;
	p_storage->items[index].data_size = ra_data_size;

	//<------------------------------------------------
	mutex_unlock(&p_storage->lock);

	return 0;
}

int process_local_ra_storage_load_and_free(
	struct process_local_ra_storage_container *p_storage, u32 *p_ra_address,
	u32 *p_ra_data_size)
{
	int result = -ENOENT;
	int index;
	const pid_t caller_pid = current->pid;

	if ((p_storage == NULL) || (p_ra_address == NULL) ||
	    (p_ra_data_size == NULL))
		return -EINVAL;

	mutex_lock(&p_storage->lock);
	//------------------------------------------------>

	// is there any entry for this pid?
	index = find_item_for_pid(p_storage, caller_pid);
	if (index >= 0) {
		*p_ra_address = p_storage->items[index].address;
		*p_ra_data_size = p_storage->items[index].data_size;
		remove_item_locked(p_storage, index);
		result = 0;
	}

	//<------------------------------------------------
	mutex_unlock(&p_storage->lock);

	if (index < 0)
		pr_warn("process_local_ra_storage_load_and_free(PID: %d) -> no entry found!\n",
			caller_pid);
	//pr_debug(
	//	"process_local_ra_storage_load_and_free(PID: %d, size: %d, address: 0x%04x, res: %d)\n",
	//	caller_pid, *p_ra_data_size, *p_ra_address, result);

	return result;
}

static int
find_free_index_locked(struct process_local_ra_storage_container *p_storage)
{
	unsigned int current_index;
	int free_index = -1;
	for (current_index = 0; current_index < ARRAY_SIZE(p_storage->items);
	     current_index++) {
		if (!p_storage->items[current_index].is_used) {
			free_index = (int)current_index;
			break;
		}
	}

	//pr_debug(" find_free_index_locked() -> Index: %d\n", free_index);
	return free_index;
}

static int
find_oldest_item_locked(struct process_local_ra_storage_container *p_storage)
{
	unsigned int current_index;
	int oldest_index = -1;
	s32 oldest_order_number = MAX_RA_STORAGE_ITEMS;
	for (current_index = 0; current_index < ARRAY_SIZE(p_storage->items);
	     current_index++) {
		const s32 current_order_number =
			p_storage->items[current_index].order_number;
		if (p_storage->items[current_index].is_used &&
		    current_order_number <= oldest_order_number) {
			oldest_index = (int)current_index;
			oldest_order_number = current_order_number;
		}
	}

	//pr_debug(" find_oldest_item_locked() -> Index: %d\n", oldest_index);
	return oldest_index;
}

static int
find_item_for_pid(struct process_local_ra_storage_container *p_storage,
		  pid_t pid)
{
	unsigned int current_index;
	int pid_index = -1;
	for (current_index = 0; current_index < ARRAY_SIZE(p_storage->items);
	     current_index++) {
		if (p_storage->items[current_index].is_used &&
		    p_storage->items[current_index].item_owner == pid) {
			pid_index = (s32)current_index;
			break;
		}
	}

	//pr_debug(" find_item_for_pid() -> Index: %d\n", pid_index);
	return pid_index;
}

static s32 find_largest_order_number_locked(
	struct process_local_ra_storage_container *p_storage)
{
	unsigned int current_index;
	s32 largest_order_number = -1;
	for (current_index = 0; current_index < ARRAY_SIZE(p_storage->items);
	     current_index++) {
		const s32 current_order_number =
			p_storage->items[current_index].order_number;
		if (p_storage->items[current_index].is_used &&
		    current_order_number > largest_order_number) {
			largest_order_number = current_order_number;
		}
	}

	//pr_debug(" find_largest_order_number_locked() -> OrderNr: %d\n",
	//	 largest_order_number);
	return largest_order_number;
}

static void remove_items_for_pid_locked(
	struct process_local_ra_storage_container *p_storage, pid_t pid)
{
	while (true) {
		int index_for_pid = find_item_for_pid(p_storage, pid);
		if (index_for_pid >= 0)
			remove_item_locked(p_storage, index_for_pid);
		else
			break;
	};
}

static void
remove_item_locked(struct process_local_ra_storage_container *p_storage,
		   unsigned int item_index_to_remove)
{
	unsigned int index;
	int removed_order_number;
	//pr_debug(" remove_item_locked(Index: %d)\n", item_index_to_remove);

	// remove the item
	removed_order_number =
		p_storage->items[item_index_to_remove].order_number;
	p_storage->items[item_index_to_remove].is_used = 0;

	// and close the gap by moving all newer elements up by one.
	for (index = 0; index < ARRAY_SIZE(p_storage->items); index++) {
		if (p_storage->items[index].is_used &&
		    p_storage->items[index].order_number >
			    removed_order_number) {
			p_storage->items[index].order_number -= 1;
		}
	}
}

__attribute__((unused)) static void
dump_storage_locked(struct process_local_ra_storage_container *p_storage)
{
	unsigned int index;
	pr_debug("dump_storage_locked()\n");
	for (index = 0; index < ARRAY_SIZE(p_storage->items); index++) {
		pr_debug(
			" [%d] Used: %d, OrderNr: %d, Owner: %d, Size: %d, Address: 0x%04x\n",
			index, p_storage->items[index].is_used,
			p_storage->items[index].order_number,
			p_storage->items[index].item_owner,
			p_storage->items[index].data_size,
			p_storage->items[index].address);
	}
}