// SPDX-License-Identifier: GPL-2.0

#include <linux/kernel.h>

#include "basler-camera-driver.h"
#include "device.h"
#include "register-io.h"
#include "memoryXX.h"

/*
 * ABRM register offsets
 *
 */
#define ABRM_GENCP_VERSION 0x0
#define ABRM_MANUFACTURER_NAME 0x4
#define ABRM_MODEL_NAME 0x44
#define ABRM_FAMILY_NAME 0x84
#define ABRM_DEVICE_VERSION 0xC4
#define ABRM_MANUFACTURER_INFO 0x104
#define ABRM_SERIAL_NUMBER 0x144
#define ABRM_USER_DEFINED_NAME 0x184
#define ABRM_DEVICE_CAPABILITIES 0x1C4

/*
 * ABRM register bits
 *
 */
#define ABRM_DEVICE_CAPABILITIES_USER_DEFINED_NAMES_SUPPORT 0x1
#define ABRM_DEVICE_CAPABILITIES_STRING_ENCODING 0x0f
#define ABRM_DEVICE_CAPABILITIES_FAMILY_NAME 0x100

/*
 * Maximum read i2c burst
 *
 * TODO: To be replace by a register call of SBRM
 *
 */
#define I2C_MAXIMUM_READ_BURST 8

static int basler_read_register(struct basler_camera_dev *sensor, u8 *buffer,
				u8 buffer_size, u16 register_address)
{
	int ret = 0;
	u8 bytes_done = 0;
	struct i2c_client *client = sensor->i2c_client;

	do {
		u8 next_transfer_size =
			(u16)min(I2C_MAXIMUM_READ_BURST,
				 ((int)buffer_size - bytes_done));
		if (atomic_read(&sensor->memory_address_size) ==
		    BASLER_MEMROY_ADDRESS_SIZE_16BIT) {
			__be16 l_register_address =
				cpu_to_be16(register_address + bytes_done);
			ret = basler_read_memory_raw_16bit(
				client, (u8 *)buffer + bytes_done,
				next_transfer_size, l_register_address);
		} else if (atomic_read(&sensor->memory_address_size) ==
			   BASLER_MEMROY_ADDRESS_SIZE_32BIT) {
			__be32 l_register_address =
				cpu_to_be32(register_address + bytes_done);
			ret = basler_read_memory_raw_32bit(
				client, (u8 *)buffer + bytes_done,
				next_transfer_size, l_register_address);
		} else {
			dev_err(&sensor->i2c_client->dev,
				"basler_read_register() unsupported memory access size: %d\n",
				atomic_read(&sensor->memory_address_size));
			return -EINVAL;
		}
		if (ret < 0) {
			pr_err("basler_read_register_chunk() failed: %d\n",
			       ret);
			return ret;
		}

		if (ret == 0) {
			pr_err("basler_read_register_chunk() read 0 bytes.\n");
			return -EIO;
		}

		bytes_done = bytes_done + ret;
	} while (bytes_done < buffer_size);

	return bytes_done;
}

int basler_retrieve_device_information(struct basler_camera_dev *sensor,
				       struct basler_device_information *bdi)
{
	int ret = 0;
	u64 deviceCapabilities = 0;
	u64 stringEncoding = 0;
	__be64 deviceCapabilitiesBe = 0;
	__be32 gencpVersionBe = 0;

	bdi->_magic = BDI_MAGIC;

	ret = basler_read_register(sensor, (u8 *)&deviceCapabilitiesBe,
				   sizeof(deviceCapabilitiesBe),
				   ABRM_DEVICE_CAPABILITIES);
	if (ret < 0) {
		pr_err("basler_read_register() failed: %d\n", ret);
		return ret;
	}

	if (ret != sizeof(deviceCapabilitiesBe)) {
		pr_err("basler_read_register() not read full buffer = %d bytes\n",
		       ret);
		return -EIO;
	}

	deviceCapabilities = be64_to_cpu(deviceCapabilitiesBe);
	stringEncoding = (deviceCapabilities &
			  ABRM_DEVICE_CAPABILITIES_STRING_ENCODING) >>
			 4;
	pr_debug("deviceCapabilities = 0x%llx\n", deviceCapabilities);
	pr_debug("String Encoding = 0x%llx\n", stringEncoding);

	ret = basler_read_register(sensor, (u8 *)&gencpVersionBe,
				   sizeof(gencpVersionBe), ABRM_GENCP_VERSION);
	if (ret < 0) {
		pr_err("basler_read_register() failed: %d\n", ret);
		return ret;
	}

	if (ret != sizeof(gencpVersionBe)) {
		pr_err("basler_read_register() not read full buffer = %d bytes\n",
		       ret);
		return -EIO;
	}

	bdi->gencpVersion = be32_to_cpu(gencpVersionBe);
	pr_debug("l_gencpVersion = %d.%d\n",
		 (bdi->gencpVersion & 0xffff0000) >> 16,
		 bdi->gencpVersion & 0xffff);

	ret = basler_read_register(sensor, bdi->deviceVersion,
				   GENCP_STRING_BUFFER_SIZE,
				   ABRM_DEVICE_VERSION);
	if (ret < 0) {
		pr_err("basler_read_register() failed: %d\n", ret);
		return ret;
	}

	if (ret != GENCP_STRING_BUFFER_SIZE) {
		pr_err("basler_read_register() not read full buffer = %d bytes\n",
		       ret);
		return -EIO;
	}

	pr_debug("bdi->deviceVersion = %s\n", bdi->deviceVersion);

	ret = basler_read_register(sensor, bdi->serialNumber,
				   GENCP_STRING_BUFFER_SIZE,
				   ABRM_SERIAL_NUMBER);
	if (ret < 0) {
		pr_err("basler_read_register() failed: %d\n", ret);
		return ret;
	}

	if (ret != GENCP_STRING_BUFFER_SIZE) {
		pr_err("basler_read_register() not read full buffer = %d bytes\n",
		       ret);
		return -EIO;
	}

	pr_debug("bdi->serialNumber = %s\n", bdi->serialNumber);

	ret = basler_read_register(sensor, bdi->manufacturerName,
				   GENCP_STRING_BUFFER_SIZE,
				   ABRM_MANUFACTURER_NAME);
	if (ret < 0) {
		pr_err("basler_read_register() failed: %d\n", ret);
		return ret;
	}

	if (ret != GENCP_STRING_BUFFER_SIZE) {
		pr_err("basler_read_register() not read full buffer = %d bytes\n",
		       ret);
		return -EIO;
	}

	pr_debug("bdi->manufacturerName = %s\n", bdi->manufacturerName);

	ret = basler_read_register(sensor, bdi->modelName,
				   GENCP_STRING_BUFFER_SIZE, ABRM_MODEL_NAME);
	if (ret < 0) {
		pr_err("basler_read_register() failed: %d\n", ret);
		return ret;
	}

	if (ret != GENCP_STRING_BUFFER_SIZE) {
		pr_err("basler_read_register() not read full buffer = %d bytes\n",
		       ret);
		return -EIO;
	}

	pr_debug("bdi->modelName = %s\n", bdi->modelName);

	if (deviceCapabilities & ABRM_DEVICE_CAPABILITIES_FAMILY_NAME) {
		ret = basler_read_register(sensor, bdi->familyName,
					   GENCP_STRING_BUFFER_SIZE,
					   ABRM_FAMILY_NAME);
		if (ret < 0) {
			pr_err("basler_read_register() failed: %d\n", ret);
			return ret;
		}

		if (ret != GENCP_STRING_BUFFER_SIZE) {
			pr_err("basler_read_register() not read full buffer = %d bytes\n",
			       ret);
			return -EIO;
		}

		pr_debug("bdi->familyName = %s\n", bdi->familyName);
	} else
		pr_notice("ABRM: FamilyName not supported\n");

	if (deviceCapabilities &
	    ABRM_DEVICE_CAPABILITIES_USER_DEFINED_NAMES_SUPPORT) {
		ret = basler_read_register(sensor, bdi->userDefinedName,
					   GENCP_STRING_BUFFER_SIZE,
					   ABRM_USER_DEFINED_NAME);
		if (ret < 0) {
			pr_err("basler_read_register() failed: %d\n", ret);
			return ret;
		}

		if (ret != GENCP_STRING_BUFFER_SIZE) {
			pr_err("basler_read_register() not read full buffer = %d bytes\n",
			       ret);
			return -EIO;
		}

		pr_debug("bdi->userDefinedName = %s\n", bdi->userDefinedName);
	} else
		pr_notice("ABRM: UserDefinedName not supported\n");

	ret = basler_read_register(sensor, bdi->manufacturerInfo,
				   GENCP_STRING_BUFFER_SIZE,
				   ABRM_MANUFACTURER_INFO);
	if (ret < 0) {
		pr_err("basler_read_register() failed: %d\n", ret);
		return ret;
	}

	if (ret != GENCP_STRING_BUFFER_SIZE) {
		pr_err("basler_read_register() not read full buffer = %d bytes\n",
		       ret);
		return -EIO;
	}

	pr_debug("bdi->manufacturerInfo = %s\n", bdi->manufacturerInfo);

	/*
	 * If the strings are in ASCII - print it.
	 */
	if (stringEncoding == 0) {
		pr_info("ABRM: Manufactuturer: %s, Model: %s, Device: %s, Serial: %s\n",
			bdi->manufacturerName, bdi->modelName,
			bdi->deviceVersion, bdi->serialNumber);
	}

	return 0;
}
